from ..autograd.tensor import Tensor, pad
from ..autograd import function as AF
from . import utils as U

import math
import numpy as np
from typing import (
    Optional,
    Tuple,
    List,
    Union,
)


def linear(x: Tensor, weight: Tensor, bias: Optional[Tensor] = None) -> Tensor:
    r"""
    Applies a linear transformation to the incoming datasets: :math:`y = xW + b`.
    """
    output = x @ weight
    if bias is not None:
        return output + bias
    return output


def conv2d(x: Tensor, weight: Tensor, bias: Optional[Tensor] = None,
           stride: Union[Tuple, List, int] = 1,
           padding: Union[Tuple, List, int, str] = 0,
           dilation: Union[Tuple, List, int] = 1,
           groups: int = 1,
           padding_mode: str = 'zeros'):
    """

    :param x: Tensor shape(bs, ch_i, h_i, w_i)
    :param weight: Tensor shape(ch_o, ch_i/groups, h_k, w_k)
    :param bias: Tensor optional shape(ch_o)
    :param stride: int | Tuple | List
    :param padding: int | str | Tuple | List
    :param dilation: int | Tuple | List
    :param groups: int
    :param padding_mode: str 'zeros' | 'reflect' | 'replicate' | 'circular'
    :return: Tensor shape(bs, ch_o, h_o, w_o)
    """
    assert x.dim() == 4
    bs, ch_i, h_i, w_i = x.shape
    ch_o, ch_i_g, h_k, w_k = weight.shape
    assert bias.dim() == 1

    stride_err_msg = "value of `stride` must be tuple of 2 or int"
    if type(stride) == int:
        stride = (stride, stride)
    elif type(stride) in [tuple, list]:
        assert len(stride) == 2, stride_err_msg
    else:
        raise ValueError(stride_err_msg)

    # make sure padding is ((int, int), (int,int))
    pad_err_msg = "value of `padding` must be tuple or list of 2, 2x2 or 4 or int"
    if type(padding) == int:
        padding = ((padding, padding), (padding, padding))
    elif type(padding) in [tuple, list]:
        if len(padding) == 2:
            if type(padding[0]) == int:
                padding = ((padding[0], padding[0]), padding[1])
            elif type(padding[0]) in [tuple, list]:
                assert len(padding[0]) == 2, pad_err_msg
            else:
                raise ValueError(pad_err_msg)
            if type(padding[1]) == int:
                padding = (padding[0], (padding[1], padding[1]))
            elif type(padding[1]) in [tuple, list]:
                assert len(padding[1]) == 2, pad_err_msg
            else:
                raise ValueError(pad_err_msg)
        elif len(padding) == 4:
            padding = ((padding[0], padding[1]), (padding[2], padding[3]))
        else:
            raise ValueError(pad_err_msg)
    elif type(padding) == str:
        padding = U.str2pad2d(padding, x.shape, weight.shape, stride)
    else:
        raise ValueError(pad_err_msg)

    dia_err_msg = "value of `dilation` must be tuple of 2 or int"
    if type(dilation) == int:
        dilation = (dilation, dilation)
    elif type(dilation) in [tuple, int]:
        assert len(dilation) == 2, dia_err_msg
    else:
        raise ValueError(dia_err_msg)

    # make sure groups
    assert ch_i % groups == 0 and ch_o % groups == 0

    x_pad = padding2d(x, padding, padding_mode)

    return AF.conv2d(x_pad, weight, bias, stride=stride, dilation=dilation, groups=groups)


def padding2d(x: Tensor, padding: Union[Tuple, List, int], mode: str) -> Tensor:
    """

    :param x: Tensor (bs, ch_i, h_i, w_i)
    :param padding: int | Tuple | List
    :param mode: str 'zeros' | 'reflect' | 'replicate' | 'circular'
    :return: Tensor (bs, ch_i, h_i+2*p[0], w_i+2*p[1])
    """
    # make sure padding is ((int, int), (int,int))
    pad_err_msg = "value of `padding` must be tuple or list of 2, 2x2 or 4 or int"
    if type(padding) == int:
        padding = ((padding, padding), (padding, padding))
    elif type(padding) in [tuple, list]:
        if len(padding) == 2:
            if type(padding[0]) == int:
                padding = ((padding[0], padding[0]), padding[1])
            elif type(padding[0]) in [tuple, list]:
                assert len(padding[0]) == 2, pad_err_msg
            else:
                raise ValueError(pad_err_msg)
            if type(padding[1]) == int:
                padding = (padding[0], (padding[1], padding[1]))
            elif type(padding[1]) in [tuple, list]:
                assert len(padding[1]) == 2, pad_err_msg
            else:
                raise ValueError(pad_err_msg)
        elif len(padding) == 4:
            padding = ((padding[0], padding[1]), (padding[2], padding[3]))
        else:
            raise ValueError(pad_err_msg)
    else:
        raise ValueError(pad_err_msg)

    pad_width = ((0, 0), (0, 0), *padding)
    if mode == 'zeros':
        return pad(x, pad_width)
    elif mode == 'reflect':
        return pad(x, pad_width, 'reflect')
    elif mode == 'replicate':
        return pad(x, pad_width, 'edge')
    elif mode == 'circular':
        return pad(x, pad_width, 'symmetric')
    else:
        raise ValueError("value of `mode` must be 'zeros', 'reflect', 'replicate' or 'circular'")


def max_pool2d(x: Tensor, kernel_size: Union[Tuple, List, int],
               stride: Optional[Union[Tuple, List, int]] = None,
               padding: Union[Tuple, List, int] = 0,
               dilation: Union[Tuple, List, int] = 1,
               padding_mode: str = 'zeros') -> Tensor:
    assert x.dim() == 4

    bs, ch_i, h_i, w_i = x.shape
    if type(kernel_size) == int:
        kernel_size = (ch_i, ch_i, kernel_size, kernel_size)
    elif type(kernel_size) in [tuple, list]:
        kernel_size = (ch_i, ch_i, *kernel_size)
    _, _, h_k, w_k = kernel_size

    stride_err_msg = "value of `stride` must be tuple of 2 or int"
    if stride is None:
        stride = (kernel_size[2], kernel_size[3])
    elif type(stride) == int:
        stride = (stride, stride)
    elif type(stride) in [tuple, list]:
        assert len(stride) == 2, stride_err_msg
    else:
        raise ValueError(stride_err_msg)

    # make sure padding is ((int, int), (int,int))
    pad_err_msg = "value of `padding` must be tuple or list of 2, 2x2 or 4 or int"
    if type(padding) == int:
        padding = ((padding, padding), (padding, padding))
    elif type(padding) in [tuple, list]:
        if len(padding) == 2:
            if type(padding[0]) == int:
                padding = ((padding[0], padding[0]), padding[1])
            elif type(padding[0]) in [tuple, list]:
                assert len(padding[0]) == 2, pad_err_msg
            else:
                raise ValueError(pad_err_msg)
            if type(padding[1]) == int:
                padding = (padding[0], (padding[1], padding[1]))
            elif type(padding[1]) in [tuple, list]:
                assert len(padding[1]) == 2, pad_err_msg
            else:
                raise ValueError(pad_err_msg)
        elif len(padding) == 4:
            padding = ((padding[0], padding[1]), (padding[2], padding[3]))
        else:
            raise ValueError(pad_err_msg)
    else:
        raise ValueError(pad_err_msg)

    dia_err_msg = "value of `dilation` must be tuple of 2 or int"
    if type(dilation) == int:
        dilation = (dilation, dilation)
    elif type(dilation) in [tuple, int]:
        assert len(dilation) == 2, dia_err_msg
    else:
        raise ValueError(dia_err_msg)

    h_o = math.floor((h_i + padding[0][0] + padding[0][1] - dilation[0] * (h_k - 1) - 1) / stride[0] + 1)
    w_o = math.floor((w_i + padding[1][0] + padding[1][1] - dilation[1] * (w_k - 1) - 1) / stride[1] + 1)

    pad_inp = padding2d(x, padding, padding_mode)
    col = U.im2col2d(pad_inp, kernel_size, stride, dilation)
    a = AF.max(col, dim=-1).reshape(bs, h_o, w_o, ch_i)
    return a.permute(dims=(0, 3, 1, 2))


def dropout(x: Tensor, p: float = 0.5, training: bool = True) -> Tensor:
    return AF.dropout(x, p, training)


def flatten(x: Tensor, start_dim: int = 1, end_dim: int = -1) -> Tensor:
    return x.flatten(start_dim, end_dim)


# activation

def relu(x: Tensor) -> Tensor:
    return AF.relu(x)


def sigmoid(x: Tensor) -> Tensor:
    return AF.sigmoid(x)


def softmax(x: Tensor, dim: int = -1) -> Tensor:
    return AF.softmax(x, dim=dim)


def log_softmax(x: Tensor, dim: int = -1) -> Tensor:
    return AF.log(AF.softmax(x, dim=dim))


def tanh(x: Tensor) -> Tensor:
    return AF.tanh(x)


# loss

def l1_loss(x: Tensor, y: Tensor, reduction: str = 'mean') -> Tensor:
    # TODO: test
    L = AF.abs(x - y)

    if reduction.lower() == 'mean':
        return L.mean()
    elif reduction.lower() == 'sum':
        return L.sum()
    elif reduction.lower() == 'none':
        return L
    else:
        raise ValueError("reduction must be 'mean', 'sum', or None")


def mse_loss(x: Tensor, y: Tensor, reduction: str = 'mean') -> Tensor:
    L = (x - y) ** Tensor(2)

    if reduction.lower() == 'mean':
        return L.mean()
    elif reduction.lower() == 'sum':
        return L.sum()
    elif reduction.lower() == 'none':
        return L
    else:
        raise ValueError("reduction must be 'mean', 'sum', or None")


def binary_cross_entropy(x: Tensor, y: Tensor, weight: Optional[Tensor] = None, reduction: str = 'mean') -> Tensor:
    """
    Args:
        x: :math:`(N, *)` where :math:`*` means, any number of additional dimensions
        y: :math:`(N, *)`, same shape as the input
        weight: a manual rescaling weight given to the loss of each batch element. If given, has to be a Tensor of size N.
        reduction:

    Returns:
        scalar. If :attr:`reduction` is ``'none'``, then :math:`(N, *)`, same shape as input.
    """
    assert np.logical_and(x.data > 0, x.data < 1).all(), "elements of input should between 0 and 1"
    w = Tensor(1)
    if weight:
        w = weight

    L = -w * (y * AF.log(x) + (1 - y) * AF.log(1 - x))

    if reduction == 'mean':
        return L.mean()
    elif reduction == 'sum':
        return L.sum()
    elif reduction.lower() == 'none':
        return L
    else:
        raise ValueError("reduction must be 'mean', 'sum', or None")


def nll_loss(x: Tensor, y: Tensor, weight: Optional[Tensor] = None, reduction: str = 'mean') -> Tensor:
    # TODO: impl weight
    assert x.dim() == 2
    assert y.dim() == 1

    w = Tensor(1)
    if weight:
        assert weight.dim() == 1
        w = weight

    L = -w * AF.nll(x, y, dim=-1)

    if reduction.lower() == 'mean':
        return L.mean()
    elif reduction.lower() == 'sum':
        return L.sum()
    elif reduction.lower() == 'none':
        return L
    else:
        raise ValueError("reduction must be 'mean', 'sum', or 'none'")


def cross_entropy(x: Tensor, y: Tensor, weight=None, reduction: str = 'mean') -> Tensor:
    # TODO: impl weight
    return nll_loss(log_softmax(x), y, weight, reduction)



from cranet import Tensor

from .module import Module
from .. import functional as F

from typing import (
    Tuple,
    List,
    Union,
)


class MaxPool2d(Module):
    def __init__(self, kernel_size: Union[Tuple, List, int],
                 stride: Union[Tuple, List, int] = None,
                 padding: Union[Tuple, List, int] = 0,
                 dilation: Union[Tuple, List, int] = 1,
                 padding_mode: str = 'zeros'):
        super().__init__()

        if type(kernel_size) == int:
            self.kernel_size = (kernel_size, kernel_size)
        elif type(kernel_size) in [tuple, list]:
            self.kernel_size = kernel_size
        else:
            raise ValueError("value of `kernel_size` must be int, tuple or list")

        if stride is None:
            self.stride = self.kernel_size
        elif type(stride) == int:
            self.stride = (stride, stride)
        elif type(stride) in [tuple, list]:
            assert len(stride) == 2
            self.stride = stride
        else:
            raise ValueError("value of `stride` must be int, tuple or list")

        if type(padding) == int:
            self.padding = (padding, padding)
        elif type(padding) in [tuple, list]:
            self.padding = padding
        elif type(padding) == str:
            self.padding = padding
        else:
            raise ValueError("value of `padding` must be int, tuple, list or str")

        if type(dilation) == int:
            self.dilation = (dilation, dilation)
        elif type(dilation) in [tuple, list]:
            self.dilation = dilation
        else:
            raise ValueError("value of `dilation` must be int, tuple or list")

        self.padding_mode = padding_mode

    def forward(self, x: Tensor) -> Tensor:
        return F.max_pool2d(x, self.kernel_size, self.stride, self.padding, self.dilation, self.padding_mode)

    def __repr__(self) -> str:
        return f"MaxPool2d"



from __future__ import annotations

from cranet import Tensor

from .module import Module

from collections import OrderedDict
from typing import (
    List,
    Dict,
    Union,
)


class Sequential(Module):
    def __init__(self, *args) -> None:
        super().__init__()
        if len(args) == 1 and isinstance(args[0], OrderedDict):
            for key, module in args[0].items():
                self.register_module(key, module)
        else:
            for idx, module in enumerate(args):
                self.register_module(str(idx), module)

    def forward(self, x: Tensor) -> Tensor:
        for (_, module) in self._modules.items():
            x = module(x)
        return x

    def __len__(self) -> int:
        return len(self.layers)

    def __getitem__(self, idx: Union[int, slice]) -> Union[Module, Sequential]:
        if isinstance(idx, slice):
            return Sequential(*self.layers[idx])
        elif isinstance(idx, int):
            return self.layers[idx]
        else:
            msg = '{cls.__name__} indices must be integer or slice object'
            raise TypeError(msg.format(cls=Sequential))

    def __repr__(self) -> str:
        # TODO: impl
        name_idx: Dict[str, int] = {}
        ss: List[str] = []
        for (_, i) in self._modules.items():
            name = i.get_name()
            if name in name_idx:
                name_idx[name] += 1
                name += str(name_idx[name])
            else:
                name_idx[name] = 0
                name += str(0)
            ss.append(name + "\n")
        return "Sequential:\n" + "".join(ss)



import numpy as np

import pickle
from typing import Any


def load_pickle(filename: str):
    with open(filename, "rb") as f:
        return pickle.load(f)


def dump_pickle(obj: Any, filename: str):
    with open(filename, "wb") as f:
        pickle.dump(obj, f)



from cranet import Tensor

from .module import Module
from .. import functional as F


class Flatten(Module):
    def __init__(self, start_dim=1, end_dim=-1):
        super().__init__()
        self.start_dim = start_dim
        self.end_dim = end_dim

    def forward(self, x: Tensor) -> Tensor:
        return F.flatten(x, self.start_dim, self.end_dim)

    def __repr__(self) -> str:
        return f"Flatten"



from cranet import Tensor

from .module import Module
from .. import functional as F


class Dropout(Module):
    def __init__(self, p: float = 0.5):
        super().__init__()
        self.p = p

    def forward(self, x: Tensor) -> Tensor:
        return F.dropout(x, self.p)

    def __repr__(self) -> str:
        return f"Dropout"



import numpy as np

from typing import (
    Union, Tuple, List,
)


def invert_permutation(permutation):
    if type(permutation) is tuple:
        permutation = list(permutation)
    inv = np.empty_like(permutation)
    inv[permutation] = np.arange(len(inv), dtype=inv.dtype)
    return inv


def im2col2d(x: np.ndarray,
             kernel_shape: Union[Tuple, List],
             stride: Union[Tuple, List],
             dilation: Union[Tuple, List]) -> np.ndarray:
    """

    :return (bs, h_o*w_o, ch_i, h_k*w_k)
    """
    bs, ch_i, h_i, w_i = x.shape
    _, _, h_k, w_k = kernel_shape
    h_k_t = h_k + (h_k - 1) * (dilation[0] - 1)
    w_k_t = w_k + (w_k - 1) * (dilation[1] - 1)
    ret = []
    for i in range(0, h_i - h_k_t + 1, stride[0]):
        for j in range(0, w_i - w_k_t + 1, stride[1]):
            m = x[:, :, i:i + h_k_t:dilation[0], j:j + w_k_t:dilation[1]]
            ret.append(m.reshape(bs, 1, ch_i, h_k * w_k))
    return np.concatenate(ret, axis=1)


def padding2d(x: np.ndarray,
              padding: Union[List, Tuple],
              mode: str = 'zeros') -> np.ndarray:
    """

    :param x: (bs, ch_i, h_i, w_i)
    :param padding: ((int, int), (int, int))
    :param mode: 'zeros' | 'reflect' | 'replicate' | 'circular'
    :return (bs, ch_i, h_o, w_o)
    """
    # make sure padding is ((int, int), (int,int))
    pad_err_msg = "value of `padding` must be tuple or list of 2, 2x2 or 4 or int"
    if type(padding) == int:
        padding = ((padding, padding), (padding, padding))
    elif type(padding) in [tuple, list]:
        if len(padding) == 2:
            if type(padding[0]) == int:
                padding = ((padding[0], padding[0]), padding[1])
            elif type(padding[0]) in [tuple, list]:
                assert len(padding[0]) == 2, pad_err_msg
            else:
                raise ValueError(pad_err_msg)
            if type(padding[1]) == int:
                padding = (padding[0], (padding[1], padding[1]))
            elif type(padding[1]) in [tuple, list]:
                assert len(padding[1]) == 2, pad_err_msg
            else:
                raise ValueError(pad_err_msg)
        elif len(padding) == 4:
            padding = ((padding[0], padding[1]), (padding[2], padding[3]))
        else:
            raise ValueError(pad_err_msg)
    else:
        raise ValueError(pad_err_msg)

    pad_width = ((0, 0), (0, 0), *padding)
    if mode == 'zeros':
        return np.pad(x, pad_width)
    elif mode == 'reflect':
        return np.pad(x, pad_width, 'reflect')
    elif mode == 'replicate':
        return np.pad(x, pad_width, 'edge')
    elif mode == 'circular':
        return np.pad(x, pad_width, 'symmetric')
    else:
        raise ValueError(
            "value of `mode` must be 'zeros', 'reflect', 'replicate' or 'circular'")


def str2pad2d(padding: str,
              inp_shape: Union[Tuple, List],
              kernel_shape: Union[Tuple, List],
              stride: Tuple[int, int]):
    _, _, h_i, w_i = inp_shape
    _, _, h_k, w_k = kernel_shape
    if padding.lower() == 'same':
        if h_i % stride[0] == 0:
            h_p = max(h_k - stride[0], 0)
        else:
            h_p = max(h_k - (h_i % stride[0]), 0)
        if w_i % stride[1] == 0:
            w_p = max(w_k - stride[1], 0)
        else:
            w_p = max(w_k - (w_i % stride[1]), 0)
        h_pa = h_p // 2
        h_pb = h_p - h_pa
        w_pa = w_p // 2
        w_pb = w_p - w_pa
        return (h_pa, h_pb), (w_pa, w_pb)
    elif padding.lower() == 'valid':
        return (0, 0), (0, 0)
    else:
        raise ValueError("padding string must be 'same' or 'valid'")


def rot180_2d(arr: np.ndarray) -> np.ndarray:
    bs, ch, h, w = arr.shape
    new_arr = arr.reshape((bs, ch, h * w))
    new_arr = new_arr[:, :, ::-1]
    new_arr = new_arr.reshape((bs, ch, w, h))
    return new_arr


def unstride(x: np.ndarray, stride_i) -> np.ndarray:
    bs, ch_o, h_o, w_o = x.shape
    h_st, w_st = stride_i
    if h_st == 1 and w_st == 1:
        return x
    else:
        # h_o_s = (h_o - 1) * h_st + 1
        # w_o_s = (w_o - 1) * w_st + 1
        h_o_s = h_o * h_st
        w_o_s = w_o * w_st
        x_s = np.zeros((bs, ch_o, h_o_s, w_o_s))
        for i in range(h_o):
            for j in range(w_o):
                x_s[:, :, i * h_st, j * w_st] = x[:, :, i, j]
        return x_s


def unpadding(x: np.ndarray, kernel_size) -> np.ndarray:
    _, _, h_k, w_k = kernel_size
    padding = ((h_k - 1, h_k - 1),
               (w_k - 1, w_k - 1),)
    x_p = padding2d(x, padding)
    return x_p


def pad_stride(x: np.ndarray, w_r_shape, stride_i) -> np.ndarray:
    """

    :param x: (bs, ch_o, h_o, w_o)
    :param w_r_shape: (bs, ch_i, h_k, w_k)
    :param stride_i: (int, int)
    :return:
    """
    bs, ch_o, h_o, w_o = x.shape
    _, _, h_k, w_k = w_r_shape
    h_st, w_st = stride_i

    if h_st == 1 and w_st == 1:
        x_s = x
    else:
        # h_o_s = (h_o - 1) * h_st + 1
        # w_o_s = (w_o - 1) * w_st + 1
        h_o_s = h_o * h_st
        w_o_s = w_o * w_st
        x_s = np.zeros((bs, ch_o, h_o_s, w_o_s))
        for i in range(h_o):
            for j in range(w_o):
                x_s[:, :, i * h_st, j * w_st] = x[:, :, i, j]

    padding = ((h_k - 1, h_k - 1),
               (w_k - 1, w_k - 1),)
    x_s_p = padding2d(x_s, padding)
    return x_s_p



from .activation import (
    ReLU,
    Sigmoid,
    Tanh,
    Softmax,
)
from .container import (
    Sequential
)
from .conv import (
    Conv2d
)
from .dropout import (
    Dropout
)
from .flatten import (
    Flatten
)
from .linear import (
    Linear
)
from .loss import (
    MSELoss,
    L1Loss,
    CrossEntropyLoss,
    NLLLoss,
)
from .module import (
    Module
)
from .pooling import (
    MaxPool2d
)



from __future__ import annotations

from .dataset import Dataset
from .sampler import Sampler, RandomSampler, SequentialSampler, BatchSampler
from .collate import default_collate_fn, default_convert_fn
from .fetch import DatasetFetcher

from typing import TypeVar, Generic, Optional, Union, Iterable, Callable, List, Any, Sequence

T_co = TypeVar('T_co', covariant=True)
T = TypeVar('T')

_collate_fn_t = Callable[[List[T]], Any]


class DataLoader(Generic[T_co]):
    dataset: Dataset[T_co]
    batch_size: Optional[int]
    sampler: Union[Sampler, Iterable]

    def __init__(self, dataset: Dataset[T_co],
                 batch_size: Optional[int] = 1,
                 shuffle: bool = False,
                 sampler: Union[Sampler, Iterable, None] = None,
                 batch_sampler: Union[Sampler[Sequence], Iterable[Sequence], None] = None,
                 collate_fn: Optional[_collate_fn_t] = None,
                 drop_last: bool = False):
        self.dataset = dataset
        self.batch_size = batch_size
        if batch_sampler is not None:
            batch_size = None
            drop_last = False

        if sampler is None:
            if shuffle:
                sampler = RandomSampler(dataset)
            else:
                sampler = SequentialSampler(dataset)

        if batch_size is not None and batch_sampler is None:
            batch_sampler = BatchSampler(sampler, batch_size, drop_last)

        self.batch_size = batch_size
        self.drop_last = drop_last
        self.sampler = sampler
        self.batch_sampler = batch_sampler

        if collate_fn is None:
            if self.auto_collation:
                collate_fn = default_collate_fn
            else:
                collate_fn = default_convert_fn
        self.collate_fn = collate_fn

    def __len__(self) -> int:
        return len(self.dataset) // self.batch_size

    @property
    def auto_collation(self):
        return self.batch_sampler is not None

    @property
    def index_sampler(self):
        if self.auto_collation:
            return self.batch_sampler
        else:
            return self.sampler

    def __iter__(self):
        return DataLoaderIter(self)


class DataLoaderIter:
    def __init__(self, loader: DataLoader) -> None:
        self._dataset = loader.dataset
        self._drop_last = loader.drop_last
        self._index_sampler = loader.index_sampler
        self._auto_collation = loader.auto_collation
        self._collate_fn = loader.collate_fn
        self._sampler_iter = iter(self._index_sampler)
        self._num_yielded = 0
        self._dataset_fetcher = DatasetFetcher(self._dataset, self._auto_collation, self._collate_fn, self._drop_last)

    def _reset(self):
        self._sampler_iter = iter(self._index_sampler)
        self._num_yielded = 0
        self._dataset_iter = iter(self._dataset)

    def __iter__(self):
        return self

    def _next_index(self):
        return next(self._sampler_iter)

    def _next_data(self):
        index = self._next_index()
        data = self._dataset_fetcher.fetch(index)
        return data

    def __next__(self):
        if self._sampler_iter is None:
            self._reset()
        data = self._next_data()
        self._num_yielded += 1
        return data

    def __len__(self) -> int:
        return len(self._index_sampler)



from cranet import Tensor

from .module import Module
from .. import functional as F
from ..parameter import Parameter
from .. import init

import numpy as np

from typing import (
    Optional,
)


class Linear(Module):
    r"""
    Applies a linear transformation to the incoming datasets: :math:`y = xW + b`.
    """
    __constants__ = ['in_features', 'out_features']
    in_features: int
    out_features: int
    weight: Parameter
    bias: Optional[Parameter]

    def __init__(self, in_features: int, out_features: int, bias: bool = True) -> None:
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.use_bias = bias
        self.reset_parameters()

    def reset_parameters(self) -> None:
        """init(reset) parameters"""
        k = np.sqrt(1 / self.in_features)
        weight_data = init.uniform_([self.in_features, self.out_features], -k, k)
        self.weight = Parameter(weight_data)
        if self.use_bias:
            bias_data = init.uniform_([self.out_features], -k, k)
            self.bias = Parameter(bias_data)
        else:
            self.bias = None

    def forward(self, x: Tensor) -> Tensor:
        return F.linear(x, self.weight, self.bias)

    def __repr__(self) -> str:
        return f"Linear in_features={self.in_features}, out_features={self.out_features}, bias={self.bias is not None}"



from __future__ import annotations

from .utils import invert_permutation

import numpy as np

from typing import (
    List,
    Tuple,
    Union,
    Sequence,
    Callable,
    Optional,
    NamedTuple,
    Dict,
)


class Dependency(NamedTuple):
    tensor: Tensor
    grad_fn: Callable[[np.ndarray, Optional[Dict]], np.ndarray]
    meta: Optional[Dict] = None


Shapable = Union[Tuple[int], List[int], int]
Arrayable = Union[float, list, np.ndarray]
Tensorable = Union['Tensor', float, np.ndarray]


def ensure_array(arrayable: Arrayable) -> np.ndarray:
    if isinstance(arrayable, np.ndarray):
        return arrayable
    else:
        return np.array(arrayable)


def ensure_tensor(tensorable: Tensorable) -> Tensor:
    if isinstance(tensorable, Tensor):
        return tensorable
    else:
        return Tensor(tensorable)


class Tensor:
    def __init__(self, data: Arrayable,
                 requires_grad: bool = False,
                 dependencies: List[Dependency] = None) -> None:
        self._data = ensure_array(data)
        self.requires_grad = requires_grad
        self.dependencies = dependencies or []
        self.shape = self.data.shape
        self.grad: Optional[Tensor] = None

        if self.requires_grad:
            self.zero_grad()

    @property
    def data(self) -> np.ndarray:
        return self._data

    @data.setter
    def data(self, new: np.ndarray) -> None:
        self._data = new
        # detach gradient if set new datasets
        self.grad = None

    def numpy(self) -> np.ndarray:
        return self.data

    def item(self):
        if self.shape == ():
            return self.data
        elif self.shape == (1,):
            return self.data[0]
        else:
            raise RuntimeError('must be one element')

    def zero_grad(self) -> None:
        self.grad = Tensor(np.zeros_like(self.data))

    def backward(self, grad: Optional[Tensor] = None) -> None:
        assert self.requires_grad, "called backward on a non-requires-grad tensor"
        assert self.grad is not None, "must call zero_grad before backward"

        if grad is None:
            if self.shape == ():
                grad = Tensor(1.)
            else:
                raise RuntimeError(
                    "grad must be specified for non-zero-tensor")
        self.grad.data += grad.data

        for dependency in self.dependencies:
            backward_grad = dependency.grad_fn(grad.data, dependency.meta)
            dependency.tensor.backward(Tensor(backward_grad))

    def dim(self) -> int:
        """Number of self.datasets dimensions"""
        return self._data.ndim

    def numel(self) -> int:
        """the product of the self.datasets’s dimensions."""
        return self._data.size

    def sum(self, dim: Optional[Shapable] = None) -> Tensor:
        return sum(self, dim)

    def mean(self, dim: Optional[Shapable] = None) -> Tensor:
        return mean(self, dim)

    def transpose(self, dim1: int, dim2: int) -> Tensor:
        return transpose(self, dim1, dim2)

    def permute(self, dims: Union[List, Tuple]) -> Tensor:
        return permute(self, dims)

    def reshape(self, *sp: int) -> Tensor:
        return reshape(self, sp)

    def flatten(self, start_dim=0, end_dim=-1):
        return flatten(self, start_dim, end_dim)

    def diag(self, diagonal: int = 0) -> Tensor:
        return diag(self, diagonal)

    def index_select(self, dim: int, index) -> Tensor:
        return index_select(self, dim, index)

    @property
    def T(self) -> Tensor:
        return permute(self, list(range(len(self.shape) - 1, -1, -1)))

    def detach(self) -> Tensor:
        new_data = self._data.copy()
        new_requires_grad = self.requires_grad
        return Tensor(new_data, new_requires_grad)

    def __add__(self, other) -> Tensor:
        """called if `self + other`"""
        return add(self, ensure_tensor(other))

    def __radd__(self, other) -> Tensor:
        """called if `other + self`"""
        return add(ensure_tensor(other), self)

    def __iadd__(self, other) -> Tensor:
        """called if `self += other`"""
        self.data = self.data + ensure_tensor(other).data
        return self

    def __sub__(self, other) -> Tensor:
        """called if `self - other`"""
        return sub(self, ensure_tensor(other))

    def __rsub__(self, other) -> Tensor:
        """called if `other - self`"""
        return sub(ensure_tensor(other), self)

    def __isub__(self, other) -> Tensor:
        """called if `self -= other`"""
        self.data = self.data - ensure_tensor(other).data
        return self

    def __neg__(self) -> Tensor:
        """called if `-self`"""
        return neg(self)

    def __mul__(self, other) -> Tensor:
        """called if `self * other`"""
        return mul(self, ensure_tensor(other))

    def __rmul__(self, other) -> Tensor:
        """called if `other - self`"""
        return mul(ensure_tensor(other), self)

    def __imul__(self, other) -> Tensor:
        """called if `self *= other`"""
        self.data = self.data * ensure_tensor(other).data
        return self

    def __truediv__(self, other) -> Tensor:
        """called if `self / other`"""
        return truediv(self, ensure_tensor(other))

    def __rtruediv__(self, other) -> Tensor:
        """called if `other / self`"""
        return truediv(ensure_tensor(other), self)

    def __itruediv__(self, other) -> Tensor:
        """called if `self /= other`"""
        self.data = self.data / ensure_tensor(other).data
        return self

    def __matmul__(self, other) -> Tensor:
        """called if `self @ other`"""
        return matmul(self, ensure_tensor(other))

    def __rmatmul__(self, other) -> Tensor:
        """called if `other @ self`"""
        return matmul(ensure_tensor(other), self)

    def __imatmul__(self, other) -> Tensor:
        """called if `self @= other`"""
        self.data = self.data @ ensure_tensor(other).data
        return self

    def __pow__(self, e: Tensor) -> Tensor:
        """called if `self ** other`"""
        return power(self, ensure_tensor(e))

    def __eq__(self, other: object) -> bool:
        """test if `self == other`"""
        if not isinstance(other, Tensor):
            return NotImplemented
        if self.data.shape != other.data.shape:
            return False
        return (self.data == other.data).all()

    def __getitem__(self, idxs):
        return _slice(self, idxs)

    def __repr__(self) -> str:
        return f"Tensor({self.data}, shape={self.shape}, requires_grad={self.requires_grad})"

    def __hash__(self) -> int:
        return hash(str(self.data.data))


def tensor(data: Arrayable, requires_grad: bool = False, dtype=None) -> Tensor:
    if type(data).__module__ == np.__name__:
        return Tensor(data.copy(), requires_grad=requires_grad)
    return Tensor(data, requires_grad=requires_grad)


def as_tensor(data: Arrayable, requires_grad: bool = False) -> Tensor:
    return Tensor(data, requires_grad=requires_grad)


def stack(tensors: List[Tensor], dim: int = 0) -> Tensor:
    tensors = [unsqueeze(i, dim) for i in tensors]
    return concat(tensors, dim)


def zeros(shape, dtype=None, order='C', requires_grad=False) -> Tensor:
    return Tensor(np.zeros(shape=shape, dtype=dtype, order=order), requires_grad=requires_grad)


def zeros_like(a, dtype=None, order='K', subok=True, shape=None, requires_grad=False) -> Tensor:
    if dtype is None:
        dtype = a.data.dtype
    return Tensor(np.zeros_like(a=a, dtype=dtype, order=order, subok=subok, shape=shape), requires_grad=requires_grad)


def ones(shape, dtype=None, order='C', requires_grad=False) -> Tensor:
    return Tensor(np.ones(shape=shape, dtype=dtype, order=order), requires_grad=requires_grad)


def ones_like(a: Tensor, dtype=None, requires_grad=False) -> Tensor:
    if dtype is None:
        dtype = a.data.dtype
    return Tensor(np.ones_like(a=a, dtype=dtype, shape=a.shape), requires_grad=requires_grad)


def _slice(t: Tensor, idxs) -> Tensor:
    data = t.data[idxs]
    requires_grad = t.requires_grad
    dependencies: List[Dependency] = []

    if requires_grad:
        def grad_fn(grad: np.ndarray, _) -> np.ndarray:
            ret = np.zeros_like(t.data)
            ret[idxs] = grad
            assert ret.shape == t.shape
            return ret

        dependencies.append(Dependency(t, grad_fn, meta={"name": "slice"}))

    return Tensor(data, requires_grad, dependencies)


def concat(tensors: Sequence[Tensor], dim: int = 0) -> Tensor:
    assert len(tensors) > 0
    i_dim = tensors[0].dim()
    data = np.concatenate([t.data for t in tensors], axis=dim)
    requires_grad = True in [t.requires_grad for t in tensors]
    dependencies: List[Dependency] = []

    a = 0
    b = 0
    for i, t in enumerate(tensors):
        b += tensors[i].shape[dim]
        if t.requires_grad:
            def grad_fn(grad: np.ndarray, meta) -> np.ndarray:
                idx = tuple([np.s_[meta["a"]:meta["b"]] if j == dim else np.s_[:] for j in range(i_dim)])
                return grad[idx]

            dependencies.append(Dependency(t, grad_fn, meta={"name": "concat", "id": i, "a": a, "b": b}))
        a += tensors[i].shape[dim]

    return Tensor(data, requires_grad, dependencies)


def unsqueeze(x: Tensor, dim: int) -> Tensor:
    data = np.expand_dims(x.data, axis=dim)
    requires_grad = x.requires_grad
    dependencies = []

    if x.requires_grad:
        def grad_fn(grad: np.ndarray, _) -> np.ndarray:
            return grad.reshape(x.shape)

        dependencies.append(Dependency(x, grad_fn, meta={"name": "unsqueeze"}))

    return Tensor(data, requires_grad, dependencies)


def reshape(t: Tensor, shape: Shapable) -> Tensor:
    data = t.data.reshape(shape)
    requires_grad = t.requires_grad
    dependencies: List[Dependency] = []

    if t.requires_grad:
        def grad_fn(grad: np.ndarray, _) -> np.ndarray:
            return grad.reshape(t.shape)

        dependencies.append(Dependency(t, grad_fn, meta={"name": "reshape"}))

    return Tensor(data, requires_grad, dependencies)


def flatten(t: Tensor, start_dim=0, end_dim=-1) -> Tensor:
    if start_dim < 0:
        start_dim = t.data.ndim + start_dim
    if end_dim < 0:
        end_dim = t.data.ndim + end_dim
    sp = []
    for i in range(0, start_dim):
        sp.append(t.shape[i])
    p = 1
    for i in range(start_dim, end_dim + 1):
        p *= t.shape[i]
    sp.append(p)
    for i in range(end_dim + 1, t.data.ndim):
        sp.append(t.shape[i])
    data = t.data.reshape(sp)
    requires_grad = t.requires_grad
    dependencies: List[Dependency] = []

    if t.requires_grad:
        def grad_fn(grad: np.ndarray, _) -> np.ndarray:
            return grad.reshape(t.shape)

        dependencies.append(Dependency(t, grad_fn, meta={"name": "flatten"}))

    return Tensor(data, requires_grad, dependencies)


def pad(t: Tensor, pad_width, mode='constant', **kwargs) -> Tensor:
    data = np.pad(t.data, pad_width, mode, **kwargs)
    requires_grad = t.requires_grad
    dependencies: List[Dependency] = []

    if t.requires_grad:
        def grad_fn(grad: np.ndarray, _) -> np.ndarray:
            idx = tuple([np.s_[pad_width[i][0]:grad.shape[i] - pad_width[i][1]] for i in range(grad.ndim)])
            return grad[idx]

        dependencies.append(Dependency(t, grad_fn, meta={"name": "pad"}))

    return Tensor(data, requires_grad, dependencies)


def index_select(t: Tensor, dim: int, index: List[int]) -> Tensor:
    # TODO: impl grad
    data = np.take(t.data, index, dim)
    requires_grad = t.requires_grad
    dependencies: List[Dependency] = []

    if requires_grad:
        def grad_fn(grad: np.ndarray, _) -> np.ndarray:
            pass

        dependencies.append(Dependency(t, grad_fn, meta={"name": "index_select"}))

    return Tensor(data, requires_grad, dependencies)


def diag(t: Tensor, diagonal: int = 0) -> Tensor:
    # TODO: impl grad
    data = np.diag(t.data, diagonal)
    requires_grad = t.requires_grad
    dependencies: List[Dependency] = []

    if requires_grad:
        def grad_fn(grad: np.ndarray, _) -> np.ndarray:
            pass

        dependencies.append(Dependency(t, grad_fn, meta={"name": "diag"}))

    return Tensor(data, requires_grad, dependencies)


def sum(t: Tensor, dim: Optional[Shapable] = None) -> Tensor:
    """
    Takes a tensor and return the sum of its components
    """
    data = t.data.sum(axis=dim)
    requires_grad = t.requires_grad
    dependencies: List[Dependency] = []

    if requires_grad:
        if dim is None:
            def grad_fn(grad: np.ndarray, _) -> np.ndarray:
                return grad * np.ones_like(t.data)
        else:
            def grad_fn(grad: np.ndarray, _) -> np.ndarray:
                return np.expand_dims(grad, axis=dim) * np.ones_like(t.data)

        dependencies.append(Dependency(t, grad_fn, meta={"name": "sum"}))

    return Tensor(data, requires_grad, dependencies)


def mean(t: Tensor, dim: Optional[Shapable] = None) -> Tensor:
    data = t.sum(dim=dim)
    if dim is None:
        numel = t.numel()
    else:
        total_numel = t.numel()
        numel = total_numel // data.numel()
    return data / numel


def add(t1: Tensor, t2: Tensor) -> Tensor:
    data = np.add(t1.data, t2.data)
    requires_grad = t1.requires_grad or t2.requires_grad
    dependencies: List[Dependency] = []

    if t1.requires_grad:
        def grad_fn1(grad: np.ndarray, _) -> np.ndarray:
            ndims_added = grad.ndim - t1.data.ndim
            for _ in range(ndims_added):
                grad = grad.sum(axis=0)
            for i, dim in enumerate(t1.shape):
                if dim == 1:
                    grad = grad.sum(axis=i, keepdims=True)
            return grad

        dependencies.append(Dependency(t1, grad_fn1, meta={"name": "add_lhs"}))

    if t2.requires_grad:
        def grad_fn2(grad: np.ndarray, _) -> np.ndarray:
            ndims_added = grad.ndim - t2.data.ndim
            for _ in range(ndims_added):
                grad = grad.sum(axis=0)
            for i, dim in enumerate(t2.shape):
                if dim == 1:
                    grad = grad.sum(axis=i, keepdims=True)
            return grad

        dependencies.append(Dependency(t2, grad_fn2, meta={"name": "add_rhs"}))

    return Tensor(data, requires_grad, dependencies)


def sub(t1: Tensor, t2: Tensor) -> Tensor:
    data = t1.data - t2.data
    requires_grad = t1.requires_grad or t2.requires_grad
    dependencies: List[Dependency] = []

    if t1.requires_grad:
        def grad_fn1(grad: np.ndarray, _) -> np.ndarray:
            ndims_added = grad.ndim - t1.data.ndim
            for _ in range(ndims_added):
                grad = grad.sum(axis=0)
            for i, dim in enumerate(t1.shape):
                if dim == 1:
                    grad = grad.sum(axis=i, keepdims=True)
            return grad

        dependencies.append(Dependency(t1, grad_fn1, meta={"name": "sub_lhs"}))

    if t2.requires_grad:
        def grad_fn2(grad: np.ndarray, _) -> np.ndarray:
            ndims_added = grad.ndim - t2.data.ndim
            for _ in range(ndims_added):
                grad = grad.sum(axis=0)
            for i, dim in enumerate(t2.shape):
                if dim == 1:
                    grad = grad.sum(axis=i, keepdims=True)
            return -grad

        dependencies.append(Dependency(t2, grad_fn2, meta={"name": "sub_rhs"}))

    return Tensor(data, requires_grad, dependencies)


def neg(t: Tensor) -> Tensor:
    data = np.negative(t.data)
    requires_grad = t.requires_grad
    dependencies: List[Dependency] = []

    if requires_grad:
        dependencies.append(Dependency(t, lambda x, _: -x, meta={"name": "neg"}))

    return Tensor(data, requires_grad, dependencies)


def mul(t1: Tensor, t2: Tensor) -> Tensor:
    data = t1.data * t2.data
    requires_grad = t1.requires_grad or t2.requires_grad
    dependencies: List[Dependency] = []

    if t1.requires_grad:
        def grad_fn1(grad: np.ndarray, _) -> np.ndarray:
            grad = grad * t2.data
            ndims_added = grad.ndim - t1.data.ndim
            for _ in range(ndims_added):
                grad = grad.sum(dim=0)
            for i, dim in enumerate(t1.shape):
                if dim == 1:
                    grad = grad.sum(dim=i, keepdims=True)
            return grad

        dependencies.append(Dependency(t1, grad_fn1, meta={"name": "mul_lhs"}))

    if t2.requires_grad:
        def grad_fn2(grad: np.ndarray, _) -> np.ndarray:
            grad = grad * t1.data
            ndims_added = grad.ndim - t2.data.ndim
            for _ in range(ndims_added):
                grad = grad.sum(dim=0)
            for i, dim in enumerate(t2.shape):
                if dim == 1:
                    grad = grad.sum(dim=i, keepdims=True)
            return grad

        dependencies.append(Dependency(t2, grad_fn2, meta={"name": "mul_rhs"}))

    return Tensor(data, requires_grad, dependencies)


def truediv(t1: Tensor, t2: Tensor) -> Tensor:
    # TODO: fix epsilon
    data = np.divide(t1.data, t2.data)
    requires_grad = t1.requires_grad or t2.requires_grad
    dependencies: List[Dependency] = []

    if t1.requires_grad:
        def grad_fn1(grad: np.ndarray, _) -> np.ndarray:
            grad = np.divide(grad, t2.data)
            ndims_added = grad.ndim - t1.data.ndim
            for _ in range(ndims_added):
                grad = grad.sum(dim=0)
            for i, dim in enumerate(t1.shape):
                if dim == 1:
                    grad = grad.sum(dim=i, keepdims=True)
            return grad

        dependencies.append(Dependency(t1, grad_fn1, meta={"name": "truediv_lhs"}))

    if t2.requires_grad:
        def grad_fn2(grad: np.ndarray, _) -> np.ndarray:
            grad = - (grad / t2.data) * (t1.data / t2.data)
            # grad = np.negative(np.divide(np.multiply(grad, t1.datasets), t2.datasets * t2.datasets))
            # grad = np.negative(np.divide(np.multiply(grad, t1.datasets), np.power(t2.datasets, 2)))
            ndims_added = grad.ndim - t2.data.ndim
            for _ in range(ndims_added):
                grad = grad.sum(dim=0)
            for i, dim in enumerate(t2.shape):
                if dim == 1:
                    grad = grad.sum(dim=i, keepdims=True)
            return grad

        dependencies.append(Dependency(t2, grad_fn2, meta={"name": "truediv_rhs"}))

    return Tensor(data, requires_grad, dependencies)


def matmul(t1: Tensor, t2: Tensor) -> Tensor:
    """
    if t1 is (n1, m1) and t2 is (m1, m2), then t1 @ t2 is (n1, m2)
    so grad3 is (n1, m2)
    if t3 = t1 @ t2, and grad3 is the gradient of some function wrt t3, then
        grad1 = grad3 @ t2.T
        grad2 = t1.T @ grad3
    """
    data = np.matmul(t1.data, t2.data)
    requires_grad = t1.requires_grad or t2.requires_grad
    dependencies: List[Dependency] = []

    if t1.requires_grad:
        def grad_fn1(grad: np.ndarray, _) -> np.ndarray:
            data_t1 = t1.data
            data_t2 = t2.data
            if t1.dim() == 1:
                data_t1 = np.expand_dims(data_t1, 0)
                grad = np.expand_dims(grad, -2)
            if t2.dim() == 1:
                data_t2 = np.expand_dims(data_t2, -1)
                grad = np.expand_dims(grad, -1)
            data_t2 = np.swapaxes(data_t2, -1, -2)
            grad_t1 = grad @ data_t2
            ndims_added = grad_t1.ndim - t1.data.ndim
            for _ in range(ndims_added):
                grad_t1 = grad_t1.sum(axis=0)
            for i, dim in enumerate(t1.shape):
                if dim == 1:
                    grad_t1 = grad_t1.sum(axis=i, keepdims=True)
            assert grad_t1.shape == t1.shape
            return grad_t1

        dependencies.append(Dependency(t1, grad_fn1, meta={"name": "matmul_lhs"}))

    if t2.requires_grad:
        def grad_fn2(grad: np.ndarray, _) -> np.ndarray:
            data_t1 = t1.data
            data_t2 = t2.data
            if t1.dim() == 1:
                data_t1 = np.expand_dims(data_t1, 0)
                grad = np.expand_dims(grad, -2)
            if t2.dim() == 1:
                data_t2 = np.expand_dims(data_t2, -1)
                grad = np.expand_dims(grad, -1)
            data_t1 = np.swapaxes(data_t1, -1, -2)
            grad_t2 = data_t1 @ grad
            if t2.dim() == 1:
                grad_t2 = np.squeeze(grad_t2, axis=-1)
            ndims_added = grad_t2.ndim - t2.data.ndim
            for _ in range(ndims_added):
                grad_t2 = grad_t2.sum(axis=0)
            for i, dim in enumerate(t2.shape):
                if dim == 1:
                    grad_t2 = grad_t2.sum(axis=i, keepdims=True)
            assert grad_t2.shape == t2.shape
            return grad_t2

        dependencies.append(Dependency(t2, grad_fn2, meta={"name": "matmul_rhs"}))

    return Tensor(data, requires_grad, dependencies)


def power(x: Tensor, e: Tensor) -> Tensor:
    # TODO: fix epsilon
    data = np.power(x.data, e.data)
    requires_grad = x.requires_grad or e.requires_grad
    dependencies: List[Dependency] = []

    if x.requires_grad:
        def grad_fn1(grad: np.ndarray, _) -> np.ndarray:
            grad = grad * e.data * np.power(x.data, e.data - 1)
            ndims_added = grad.ndim - x.data.ndim
            for _ in range(ndims_added):
                grad = grad.sum(dim=0)
            for i, dim in enumerate(x.shape):
                if dim == 1:
                    grad = grad.sum(dim=i, keepdims=True)
            return grad

        dependencies.append(Dependency(x, grad_fn1, meta={"name": "power_lhs"}))

    if e.requires_grad:
        def grad_fn2(grad: np.ndarray, _) -> np.ndarray:
            grad = grad * data * np.log(x.data)
            ndims_added = grad.ndim - e.data.ndim
            for _ in range(ndims_added):
                grad = grad.sum(dim=0)
            for i, dim in enumerate(e.shape):
                if dim == 1:
                    grad = grad.sum(dim=i, keepdims=True)
            return grad

        dependencies.append(Dependency(e, grad_fn2, meta={"name": "power_rhs"}))

    return Tensor(data, requires_grad, dependencies)


def dot(t1: Tensor, t2: Tensor) -> Tensor:
    # TODO: impl, test
    assert t1.dim() <= 1 and t2.dim() <= 1
    data = np.dot(t1.data, t2.data)
    requires_grad = t1.requires_grad or t2.requires_grad
    dependencies: List[Dependency] = []

    return Tensor(data, requires_grad, dependencies)


def transpose(a: Tensor, dim1: int, dim2: int) -> Tensor:
    data = np.swapaxes(a.data, dim1, dim2)
    requires_grad = a.requires_grad
    dependencies: List[Dependency] = []

    if a.requires_grad:
        def grad_fn(grad: np.ndarray, _) -> np.ndarray:
            return np.swapaxes(grad, dim2, dim1)

        dependencies.append(Dependency(a, grad_fn, meta={"name": "transpose"}))

    return Tensor(data, requires_grad, dependencies)


def permute(a: Tensor, dims: Union[List, Tuple]) -> Tensor:
    data = a.data.transpose(dims)
    requires_grad = a.requires_grad
    dependencies: List[Dependency] = []

    if a.requires_grad:
        def grad_fn(grad: np.ndarray, _) -> np.ndarray:
            axes_t = invert_permutation(dims)
            return grad.transpose(axes_t)

        dependencies.append(Dependency(a, grad_fn, meta={"name": "permute"}))

    return Tensor(data, requires_grad, dependencies)



from .tensor import Tensor

import numpy as np


def normal(shape, loc=0.0, scale=1.0, requires_grad=False) -> Tensor:
    return Tensor(np.random.normal(loc=loc, scale=scale, size=shape), requires_grad=requires_grad)


def normal_like(a, loc=0.0, scale=1.0, requires_grad=False) -> Tensor:
    return Tensor(np.random.normal(loc=loc, scale=scale, size=a.shape), requires_grad=requires_grad)


def uniform(shape, low=0.0, high=1.0, requires_grad=False) -> Tensor:
    return Tensor(np.random.uniform(low=low, high=high, size=shape), requires_grad=requires_grad)


def uniform_like(a, low=0.0, high=1.0, requires_grad=False) -> Tensor:
    return Tensor(np.random.uniform(low=low, high=high, size=a.shape), requires_grad=requires_grad)



from ..nn.parameter import Parameter

from typing import Iterator


class Optimizer:
    def __init__(self, parameters: Iterator[Parameter], *args, **kwargs) -> None:
        self.parameters = list(parameters)

    def zero_grad(self) -> None:
        for parameter in self.parameters:
            parameter.zero_grad()

    def step(self, closure) -> None:
        r"""Performs a single optimization step (parameter update).

        Arguments:
            closure (callable): A closure that reevaluates the model and
                returns the loss. Optional for most optimizers.

        .. note::
            Unless otherwise specified, this function should not modify the
            ``.grad`` field of the parameters.
        """
        raise NotImplementedError



from __future__ import annotations

from cranet import Tensor

from ..parameter import Parameter

from typing import (
    Iterator,
    Iterable,
    Optional,
    List,
    Dict,
    Set,
    Union,
    Tuple,
    Callable,
)
from collections import OrderedDict


class Module:
    def __init__(self):
        self._parameters: Dict[str, Optional[Parameter]] = OrderedDict()
        self._buffers: Dict[str, Optional[Tensor]] = OrderedDict()
        self._non_persistent_buffers_set: Set[str] = set()
        self._modules: Dict[str, Optional[Module]] = OrderedDict()

    def register_buffer(self, name: str, tensor: Optional[Tensor], persistent: bool = True) -> None:
        r"""
        This is typically used to register a buffer that should not to be
        considered a model parameter. 

        Args:
            name (string): name of the buffer. The buffer can be accessed
                from this module using the given name
            tensor (Tensor): buffer to be registered.
            persistent (bool): whether the buffer is part of this module's
                :attr:`state_dict`.
        """
        if name == "":
            raise KeyError("key can't be empty string")
        if '.' in name:
            raise KeyError("key can't contain '.'")
        if hasattr(self, name) and name not in self._buffers:
            raise KeyError(f"attribute {name} already exist")

        self._buffers[name] = tensor
        if persistent:
            self._non_persistent_buffers_set.discard(name)
        else:
            self._non_persistent_buffers_set.add(name)

    def register_parameter(self, name: str, param: Optional[Parameter]) -> None:
        """add a parameter to the module"""
        if name == "":
            raise KeyError("key can't be empty string")
        if '.' in name:
            raise KeyError("key can't contain '.'")
        if hasattr(self, name) and name not in self._parameters:
            raise KeyError(f"attribute {name} already exist")

        if param is None:
            self._parameters[name] = None
        elif not isinstance(param, Parameter):
            raise TypeError("cranet.nn.Parameter or None required")
        else:
            self._parameters[name] = param

    def register_module(self, name: str, module: Optional[Module]) -> None:
        if name == "":
            raise KeyError("key can't be empty string")
        if '.' in name:
            raise KeyError("key can't contain '.'")
        if hasattr(self, name) and name not in self._modules:
            raise KeyError(f"attribute {name} already exist")
        self._modules[name] = module

    def get_name(self) -> str:
        return self.__class__.__name__

    def zero_grad(self):
        for p in self.parameters():
            p.zero_grad()

    def forward(self, *args, **kwargs) -> Tensor:
        raise NotImplementedError

    def __call__(self, *args, **kwargs) -> Tensor:
        return self.forward(*args, **kwargs)

    def __setstate__(self, state):
        self.__dict__.update(state)
        if '_non_persistent_buffers_set' not in self.__dict__:
            self._non_persistent_buffers_set = set()

    def __getattr__(self, name: str):
        if '_parameters' in self.__dict__:
            _parameters = self.__dict__['_parameters']
            if name in _parameters:
                return _parameters[name]
        if '_buffers' in self.__dict__:
            _buffers = self.__dict__['_buffers']
            if name in _buffers:
                return _buffers[name]
        if '_modules' in self.__dict__:
            modules = self.__dict__['_modules']
            if name in modules:
                return modules[name]
        raise AttributeError(f"'{type(self).__name__}' object has no attribute '{name}'")

    def __setattr__(self, name: str, value: Union[Tensor, Module]) -> None:
        def remove_from(*dicts_or_sets):
            for d in dicts_or_sets:
                if name in d:
                    if isinstance(d, dict):
                        del d[name]
                    else:
                        d.discard(name)

        params = self.__dict__.get('_parameters')
        if isinstance(value, Parameter):
            if params is None:
                raise AttributeError(
                    "cannot assign parameters before Module.__init__() call")
            remove_from(self.__dict__, self._buffers, self._modules, self._non_persistent_buffers_set)
            self.register_parameter(name, value)
        elif params is not None and name in params:
            if value is not None:
                raise TypeError(f"cannot assign '{type(value)}' as parameter '{name}' "
                                "(torch.nn.Parameter or None expected)")
            self.register_parameter(name, value)
        else:
            modules = self.__dict__.get('_modules')
            if isinstance(value, Module):
                if modules is None:
                    raise AttributeError("cannot assign module before Module.__init__() call")
                remove_from(self.__dict__, self._parameters, self._buffers, self._non_persistent_buffers_set)
                modules[name] = value
            elif modules is not None and name in modules:
                if value is not None:
                    raise TypeError(f"cannot assign '{type(value)}' as child module '{name}' "
                                    "(torch.nn.Module or None expected)")
                modules[name] = value
            else:
                buffers = self.__dict__.get('_buffers')
                if buffers is not None and name in buffers:
                    if value is not None and not isinstance(value, Tensor):
                        raise TypeError(f"cannot assign '{type(value)}' as buffer '{name}' "
                                        "(torch.Tensor or None expected)")
                    buffers[name] = value
                else:
                    object.__setattr__(self, name, value)

    def __delattr__(self, name):
        if name in self._parameters:
            del self._parameters[name]
        elif name in self._buffers:
            del self._buffers[name]
            self._non_persistent_buffers_set.discard(name)
        elif name in self._modules:
            del self._modules[name]
        else:
            object.__delattr__(self, name)

    def _named_members(self, get_members_fn: Callable[[Module], Iterable], prefix='', recurse=True) -> Iterator:
        r"""Helper method for yielding various names + members of modules."""
        memo = set()
        modules = self.named_modules(prefix=prefix) if recurse else [(prefix, self)]
        for module_prefix, module in modules:
            members = get_members_fn(module)
            for k, v in members:
                if v is None or v in memo:
                    continue
                memo.add(v)
                name = module_prefix + ('.' if module_prefix else '') + k
                yield name, v

    def named_parameters(self, prefix: str = '', recurse: bool = True) -> Iterator[Tuple[str, Parameter]]:
        memo: Set[Parameter] = set()
        modules: List[Tuple[str, Module]] = self.named_modules(prefix=prefix) if recurse else [(prefix, self)]
        for module_prefix, module in modules:
            members = module._parameters.items()
            for k, v in members:
                if v is None or v in memo:
                    continue
                memo.add(v)
                name = module_prefix + ('.' if module_prefix else '') + k
                yield name, v

    def named_modules(self, memo: Optional[Set[Module]] = None, prefix: str = '', remove_duplicate: bool = True):
        if memo is None:
            memo = set()
        if self not in memo:
            if remove_duplicate:
                memo.add(self)
            yield prefix, self
            for name, module in self._modules.items():
                if module is None:
                    continue
                submodule_prefix = prefix + ('.' if prefix else '') + name
                for m in module.named_modules(memo, submodule_prefix, remove_duplicate):
                    yield m

    def children(self) -> Iterator['Module']:
        r"""Returns an iterator over immediate children modules.

        Yields:
            Module: a child module
        """
        for name, module in self.named_children():
            yield module

    def named_children(self) -> Iterator[Tuple[str, 'Module']]:
        r"""Returns an iterator over immediate children modules, yielding both
        the name of the module as well as the module itself.

        Yields:
            (string, Module): Tuple containing a name and child module
        """
        memo = set()
        for name, module in self._modules.items():
            if module is not None and module not in memo:
                memo.add(module)
                yield name, module

    def parameters(self, recurse: bool = True) -> Iterator[Parameter]:
        for name, param in self.named_parameters(recurse=recurse):
            yield param

    def __repr__(self) -> str:
        return self.__class__.__name__



class DatasetFetcher:
    def __init__(self, dataset, auto_collation, collate_fn, drop_last):
        self.dataset = dataset
        self.auto_collation = auto_collation
        self.collate_fn = collate_fn
        self.drop_last = drop_last
        self.dataset_iter = iter(dataset)
        self.ended = False

    def fetch(self, slice):
        if self.ended:
            raise StopIteration

        if self.auto_collation:
            data = []
            for _ in slice:
                try:
                    data.append(next(self.dataset_iter))
                except StopIteration:
                    self.ended = True
                    break
            if len(data) == 0 or (self.drop_last and len(data) < len(slice)):
                raise StopIteration
        else:
            data = next(self.dataset_iter)
        return self.collate_fn(data)



from ..autograd import Tensor, concat

from typing import (
    Union, Tuple, List,
)


def im2col2d(x: Tensor, kernel_shape: Union[Tuple, List], stride: Union[Tuple, List], dilation: Union[Tuple, List]):
    """

    :param x: Tensor (bs, ch_i, h_i, w_i)
    :param kernel_shape: Tuple | List (ch_o, ch_i, h_k, w_k)
    :param stride: Tuple | List ()
    :param dilation: Tuple | List ()
    :return: Tensor (bs, *, i_ch*k_w*k_h)
    """
    bs, ch_i, h_i, w_i = x.shape
    _, _, h_k, w_k = kernel_shape
    h_k_t = h_k + (h_k - 1) * (dilation[0] - 1)
    w_k_t = w_k + (w_k - 1) * (dilation[1] - 1)
    ret = []
    for i in range(0, h_i - h_k_t + 1, stride[0]):
        for j in range(0, w_i - w_k_t + 1, stride[1]):
            m = x[:, :, i:i + h_k_t:dilation[0], j:j + w_k_t:dilation[1]]
            ret.append(m.reshape(bs, 1, ch_i, h_k * w_k))
    return concat(ret, dim=1)


def str2pad2d(padding: str, inp_shape, kernel_shape: Union[Tuple, List], stride: Tuple[int, int]):
    _, _, h_i, w_i = inp_shape
    _, _, h_k, w_k = kernel_shape
    if padding.lower() == 'same':
        if h_i % stride[0] == 0:
            h_p = max(h_k - stride[0], 0)
        else:
            h_p = max(h_k - (h_i % stride[0]), 0)
        if w_i % stride[1] == 0:
            w_p = max(w_k - stride[1], 0)
        else:
            w_p = max(w_k - (w_i % stride[1]), 0)
        h_pa = h_p // 2
        h_pb = h_p - h_pa
        w_pa = w_p // 2
        w_pb = w_p - w_pa
        return (h_pa, h_pb), (w_pa, w_pb)
    elif padding.lower() == 'valid':
        return (0, 0), (0, 0)
    else:
        raise ValueError("padding string must be 'same' or 'valid'")



from ..autograd.tensor import Tensor, Arrayable, Dependency

from typing import (
    List,
)


class Parameter(Tensor):
    def __init__(self, data: Arrayable = None, requires_grad: bool = True, dependencies: List[Dependency] = None):
        super().__init__(data, requires_grad, dependencies)

    def __repr__(self):
        return "Parameter containing:\n" + super().__repr__()



from cranet import Tensor

from typing import TypeVar, Generic, Tuple, Iterator

T_co = TypeVar('T_co', covariant=True)


class Dataset(Generic[T_co]):
    def __len__(self) -> int:
        raise NotImplementedError

    def __getitem__(self, item) -> T_co:
        raise NotImplementedError


class IterableDataset(Dataset[T_co]):
    def __iter__(self) -> Iterator[T_co]:
        raise NotImplementedError


class TensorDataset(Dataset[Tuple[Tensor, ...]]):
    tensors: Tuple[Tensor, ...]

    def __init__(self, *tensors: Tensor) -> None:
        self.tensors = tensors

    def __len__(self) -> int:
        pass

    def __getitem__(self, item):
        pass



from cranet import Tensor

from .module import Module
from .. import functional as F


class ReLU(Module):
    def forward(self, x: Tensor) -> Tensor:
        return F.relu(x)


class Sigmoid(Module):
    def forward(self, x: Tensor) -> Tensor:
        return F.sigmoid(x)


class Softmax(Module):
    def __init__(self, dim=-1):
        super().__init__()
        self.dim = dim

    def forward(self, x: Tensor) -> Tensor:
        return F.softmax(x, dim=self.dim)


class Tanh(Module):
    def forward(self, x: Tensor) -> Tensor:
        return F.tanh(x)



from ..autograd.tensor import Shapable
from typing import (
    Union,
)

import numpy as np


# TODO: impl initializers
# TODO: impl with tensor.zero_grad()

def uniform_(shape: Shapable, a: float = 0., b: float = 1.) -> np.ndarray:
    r"""Fills the input Tensor with values drawn from the uniform
    distribution.

    Args:
        shape: tuple or list which determines the shape of initialization matrix
        a: the lower bound of the uniform distribution
        b: the upper bound of the uniform distribution
    """
    return np.random.uniform(a, b, shape)


def normal_(shape: Shapable, mean: float = 0., std: float = 1.) -> np.ndarray:
    r"""Fills the input Tensor with values drawn from the normal
    distribution.

    Args:
        shape: tuple or list which determines the shape of initialization matrix
        mean: the mean of the normal distribution
        std: the standard deviation of the normal distribution
    """
    return np.random.normal(mean, std, shape)


def constant_(shape: Shapable, val) -> np.ndarray:
    r"""Fills the input Tensor with the value :math:`\text{val}`.

    Args:
        shape: tuple or list which determines the shape of initialization matrix
        val: the value to fill the tensor with

    """
    return np.full(shape, val)


def ones_(shape: Shapable) -> np.ndarray:
    r"""Fills the input Tensor with the scalar value `1`.

    Args:
        shape: tuple or list which determines the shape of initialization matrix
    """
    return np.ones(shape)


def zeros_(shape: Shapable) -> np.ndarray:
    r"""Fills the input Tensor with the scalar value `0`.

    Args:
        shape: tuple or list which determines the shape of initialization matrix
    """
    return np.zeros(shape)


def eye_(shape: Shapable) -> np.ndarray:
    r"""Fills the 2-dimensional input `Tensor` with the identity
    matrix. Preserves the identity of the inputs in `Linear` layers, where as
    many inputs are preserved as possible.

    Args:
        shape: tuple or list which determines the shape of initialization matrix
    """
    return np.eye(*shape)


def calculate_gain(nonlinearity: str, param=None) -> Union[int, float]:
    r"""Return the recommended gain value for the given nonlinearity function.
    The values are as follows:

    ================= ====================================================
    nonlinearity      gain
    ================= ====================================================
    Linear / Identity :math:`1`
    Conv{1,2,3}D      :math:`1`
    Sigmoid           :math:`1`
    Tanh              :math:`\frac{5}{3}`
    ReLU              :math:`\sqrt{2}`
    Leaky Relu        :math:`\sqrt{\frac{2}{1 + \text{negative\_slope}^2}}`
    SELU              :math:`\frac{3}{4}`
    ================= ====================================================
    """
    linear_fns = ['linear', 'conv1d', 'conv2d', 'conv3d', 'conv_transpose1d', 'conv_transpose2d', 'conv_transpose3d']
    if nonlinearity in linear_fns or nonlinearity == 'sigmoid':
        return 1
    elif nonlinearity == 'tanh':
        return 5.0 / 3
    elif nonlinearity == 'relu':
        return np.sqrt(2.0)
    elif nonlinearity == 'leaky_relu':
        if param is None:
            negative_slope = 0.01
        elif not isinstance(param, bool) and isinstance(param, int) or isinstance(param, float):
            # True/False are instances of int, hence check above
            negative_slope = param
        else:
            raise ValueError("negative_slope {} not a valid number".format(param))
        return np.sqrt(2.0 / (1 + negative_slope ** 2))
    elif nonlinearity == 'selu':
        return 3.0 / 4
    else:
        raise ValueError("Unsupported nonlinearity {}".format(nonlinearity))


def _calculate_correct_fan(shape: Shapable, mode):
    mode = mode.lower()
    valid_modes = ['fan_in', 'fan_out']
    if mode not in valid_modes:
        raise ValueError("Mode {} not supported, please use one of {}".format(mode, valid_modes))

    fan_in, fan_out = _calculate_fan_in_and_fan_out(shape)
    return fan_in if mode == 'fan_in' else fan_out


def _calculate_fan_in_and_fan_out(shape: Shapable):
    dimensions = len(shape)
    if dimensions < 2:
        raise ValueError("Fan in and fan out can not be computed for tensor with fewer than 2 dimensions")
    num_input_fmaps = shape[0]
    num_output_fmaps = shape[1]
    receptive_field_size = 1
    if dimensions > 2:
        receptive_field_size = shape[2:].numel()
    fan_in = num_input_fmaps * receptive_field_size
    fan_out = num_output_fmaps * receptive_field_size
    return fan_in, fan_out


def xavier_uniform_(shape: Shapable, gain: float = 1.) -> np.ndarray:
    fan_in, fan_out = _calculate_fan_in_and_fan_out(shape)
    std = gain * np.sqrt(2.0 / float(fan_in + fan_out))
    a = np.sqrt(3.0) * std  # Calculate uniform bounds from standard deviation
    return np.random.uniform(-a, a, shape)


def xavier_normal_(shape: Shapable, gain=1.) -> np.ndarray:
    fan_in, fan_out = _calculate_fan_in_and_fan_out(shape)
    std = gain * np.sqrt(2.0 / float(fan_in + fan_out))
    return np.random.normal(0., std, shape)


def kaiming_uniform_(shape: Shapable, a=0, mode='fan_in', nonlinearity='leaky_relu'):
    if 0 in shape:
        raise ValueError("Initializing zero-element tensors is a no-op")

    fan = _calculate_correct_fan(shape, mode)
    gain = calculate_gain(nonlinearity, a)
    std = gain / np.sqrt(fan)
    bound = np.sqrt(3.0) * std  # Calculate uniform bounds from standard deviation

    return np.random.uniform(-bound, bound, shape)


def kaiming_normal_(shape: Shapable, a=0, mode='fan_in', nonlinearity='leaky_relu'):
    if 0 in shape:
        raise ValueError("Initializing zero-element tensors is a no-op")

    fan = _calculate_correct_fan(shape, mode)
    gain = calculate_gain(nonlinearity, a)
    std = gain / np.sqrt(fan)
    bound = np.sqrt(3.0) * std  # Calculate uniform bounds from standard deviation

    return np.random.normal(0., std, shape)



from abc import ABC

from .module import Module
from .. import functional as F
from cranet import Tensor

from typing import (
    Optional,
)


class _Loss(Module, ABC):
    reduction: str

    def __init__(self, reduction: str = 'mean') -> None:
        super().__init__()
        self.reduction = reduction

    def __call__(self, inp: Tensor, tar: Tensor):
        return self.forward(inp, tar)


class _WeightedLoss(_Loss, ABC):
    def __init__(self, weight: Optional[Tensor] = None, reduction: str = 'mean') -> None:
        super().__init__(reduction)
        self.register_buffer('weight', weight)


class L1Loss(_Loss):
    __constants__ = ['reduction']

    def __init__(self, reduction: str = 'mean') -> None:
        super(L1Loss, self).__init__(reduction)

    def forward(self, x: Tensor, y: Tensor) -> Tensor:
        return F.l1_loss(x, y, reduction=self.reduction)


class MSELoss(_Loss):
    __constants__ = ['reduction']

    def __init__(self, reduction: str = 'mean') -> None:
        super().__init__(reduction)

    def forward(self, x: Tensor, y: Tensor) -> Tensor:
        return F.mse_loss(x, y, reduction=self.reduction)


class CrossEntropyLoss(_WeightedLoss):
    __constants__ = ['reduction']

    def __init__(self, weight: Optional[Tensor] = None, reduction: str = 'mean') -> None:
        super().__init__(weight, reduction)

    def forward(self, x: Tensor, y: Tensor) -> Tensor:
        return F.cross_entropy(x, y, weight=self.weight, reduction=self.reduction)


class NLLLoss(_WeightedLoss):
    __constants__ = ['reduction']

    def __init__(self, weight: Optional[Tensor] = None, reduction: str = 'mean') -> None:
        super().__init__(weight, reduction)

    def forward(self, x: Tensor, y: Tensor) -> Tensor:
        return F.cross_entropy(x, y, weight=self.weight, reduction=self.reduction)



from .function import (
    abs, max, min, log, exp, relu, sigmoid, softmax, tanh,
)
from .random import (
    normal, normal_like,
    uniform, uniform_like,
)
from .tensor import (
    Tensor,
    tensor, as_tensor, stack,
    zeros, zeros_like, ones, ones_like,
    add, sub, neg, mul, truediv, matmul, power, dot,
    sum, mean,
    transpose, permute, concat, reshape, flatten, pad
)
import numpy as np

uint8 = np.uint8
int32 = np.int32
int64 = np.int64
float32 = np.float32
float64 = np.float64



from .optimizer import Optimizer
from ..nn.parameter import Parameter

from typing import Iterator


class SGD(Optimizer):
    def __init__(self, parameters: Iterator[Parameter], lr: float = 0.01, *args, **kwargs) -> None:
        super().__init__(parameters, *args, **kwargs)
        self.lr = lr

    def step(self, **kwargs) -> None:
        for parameter in self.parameters:
            parameter -= parameter.grad * self.lr



# TODO: impl Hebb
from cranet.nn.modules import Module
from .optimizer import Optimizer
from ..nn.parameter import Parameter

from typing import Iterator


class Oja(Optimizer):
    def __init__(self, parameters: Iterator[Parameter], lr: float = 0.01, *args, **kwargs) -> None:
        super().__init__(parameters, *args, **kwargs)
        self.lr = lr

    def step(self, model: Module) -> None:
        for parameter in self.parameters:
            parameter = parameter + self.lr



from cranet import Tensor

from .module import Module
from .. import functional as F
from ..parameter import Parameter
from .. import init

import math
from typing import (
    Optional,
    Tuple,
    List,
    Union,
)


class Conv2d(Module):
    in_channels: int
    out_channels: int
    kernel_size: Union[Tuple, List, int]
    stride: Union[Tuple, List, int]
    padding: Union[Tuple, List, int, str]
    dilation: Union[Tuple, List, int]
    groups: int
    padding_mode: str
    weight: Parameter
    bias: Optional[Parameter]

    def __init__(self, in_channels: int, out_channels: int,
                 kernel_size: Union[Tuple, List, int],
                 stride: Union[Tuple, List, int] = 1,
                 padding: Union[Tuple, List, int, str] = 0,
                 dilation: Union[Tuple, List, int] = 1,
                 groups: int = 1, bias: bool = True, padding_mode: str = 'zeros'):
        """
            TODO: impl dilation, groups

            :param in_channels: str
            :param out_channels: str
            :param kernel_size: int | Tuple[int, int]
            :param stride: int | Tuple[int, int]
            :param padding: int | Tuple[int, int]
            :param dilation: int | Tuple[int, int]
            :param groups: int
            :param bias: bool
            :param padding_mode: str 'zeros'| 'reflect'| 'replicate'| 'circular'
            :return: Tensor shape(bs, ch_o, h_o, w_o)
            """
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels

        if type(kernel_size) == int:
            self.kernel_size = (kernel_size, kernel_size)
        elif type(kernel_size) in [tuple, list]:
            self.kernel_size = kernel_size
        else:
            raise ValueError("value of `kernel_size` must be int, tuple or list")

        if type(stride) == int:
            self.stride = (stride, stride)
        elif type(stride) in [tuple, list]:
            assert len(stride) == 2
            self.stride = stride
        else:
            raise ValueError("value of `stride` must be int, tuple or list")

        if type(padding) == int:
            self.padding = (padding, padding)
        elif type(padding) in [tuple, list]:
            self.padding = padding
        elif type(padding) == str:
            self.padding = padding
        else:
            raise ValueError("value of `padding` must be int, tuple, list or str")

        if type(dilation) == int:
            self.dilation = (dilation, dilation)
        elif type(dilation) in [tuple, list]:
            self.dilation = dilation
        else:
            raise ValueError("value of `dilation` must be int, tuple or list")

        self.groups = groups
        self.padding_mode = padding_mode
        self.reset_parameters(bias)

    def reset_parameters(self, use_bias: bool = True) -> None:
        """init(reset) parameters"""
        sqrt_k = math.sqrt(self.groups / (self.in_channels * self.kernel_size[0] * self.kernel_size[1]))
        weight_data = init.uniform_([self.out_channels, self.in_channels // self.groups, self.kernel_size[0], self.kernel_size[1]], -sqrt_k, sqrt_k)
        self.weight = Parameter(weight_data)
        if use_bias:
            bias_data = init.uniform_([self.out_channels], -sqrt_k, sqrt_k)
            self.bias = Parameter(bias_data)
        else:
            self.bias = None

    def forward(self, x: Tensor) -> Tensor:
        return F.conv2d(x, self.weight, self.bias, self.stride, self.padding,
                        self.dilation, self.groups, self.padding_mode)

    def __repr__(self) -> str:
        return f"Conv2d"



import random
from typing import TypeVar, Generic, Sized, Iterator, Union, Iterable, List

T_co = TypeVar('T_co', covariant=True)


class Sampler(Generic[T_co]):
    def __iter__(self) -> Iterator[T_co]:
        raise NotImplementedError


class SequentialSampler(Sampler[int]):
    data_source: Sized

    def __init__(self, data_source: Sized) -> None:
        super().__init__()
        self.data_source = data_source

    def __iter__(self) -> Iterator[int]:
        return iter(range(len(self.data_source)))

    def __len__(self) -> int:
        return len(self.data_source)


class RandomSampler(Sampler):
    data_source: Sized

    def __init__(self, data_source: Sized):
        self.data_source = data_source

    def __iter__(self) -> Iterator[int]:
        l = list(range(len(self.data_source)))
        random.shuffle(l)
        return iter(l)

    def __len__(self) -> int:
        return len(self.data_source)


class BatchSampler(Sampler):
    def __init__(self, sampler: Union[Sampler[int], Iterable[int]], batch_size: int, drop_last: bool) -> None:
        if not isinstance(batch_size, int) or isinstance(batch_size, bool) or \
                batch_size <= 0:
            raise ValueError("batch_size should be a positive integer value, "
                             "but got batch_size={}".format(batch_size))
        if not isinstance(drop_last, bool):
            raise ValueError("drop_last should be a boolean value, but got "
                             "drop_last={}".format(drop_last))
        self.sampler = sampler
        self.batch_size = batch_size
        self.drop_last = drop_last

    def __iter__(self) -> Iterator[List[int]]:
        batch = []
        for idx in self.sampler:
            batch.append(idx)
            if len(batch) == self.batch_size:
                yield batch
                batch = []
        if len(batch) > 0 and not self.drop_last:
            yield batch

    def __len__(self) -> int:
        if self.drop_last:
            return len(self.sampler) // self.batch_size  # type: ignore[arg-type]
        else:
            return (len(self.sampler) + self.batch_size - 1) // self.batch_size  # type: ignore[arg-type]



import cranet
from collections import Sequence, Mapping


def default_collate_fn(batch):
    elm = batch[0]
    elm_type = type(elm)
    if isinstance(elm, cranet.Tensor):
        return cranet.stack(batch, 0)
    elif isinstance(elm, float):
        return cranet.tensor(batch, dtype=cranet.float64)
    elif isinstance(elm, int):
        return cranet.tensor(batch)
    elif isinstance(elm, str):
        return
    elif isinstance(elm, Mapping):
        return {key: default_collate_fn([d[key] for d in batch]) for key in elm}
    elif isinstance(elm, tuple) and hasattr(elm, '_fields'):  # namedtuple
        return elm_type(*(default_collate_fn(samples) for samples in zip(*batch)))
    elif isinstance(elm, Sequence):
        # check to make sure that the elements in batch have consistent size
        it = iter(batch)
        elem_size = len(next(it))
        if not all(len(elem) == elem_size for elem in it):
            raise RuntimeError('each element in list of batch should be of equal size')
        transposed = zip(*batch)
        return [default_collate_fn(samples) for samples in transposed]
    raise TypeError


def default_convert_fn():
    pass



from __future__ import annotations

from .tensor import Tensor, Dependency
from . import utils as U

import numpy as np

import math
from typing import (
    Optional, Union, List, Tuple
)


def abs(x: Tensor) -> Tensor:
    data = np.absolute(x.data)
    requires_grad = x.requires_grad
    dependencies = []

    if x.requires_grad:
        def grad_fn(grad: np.ndarray, _) -> np.ndarray:
            return grad * np.where(x.data > 0, 1, -1)

        dependencies.append(Dependency(x, grad_fn, meta={"name": "abs"}))

    return Tensor(data, requires_grad, dependencies)


def max(x: Tensor, dim=None, keepdim=False) -> Tensor:
    # TODO: fix epsilon
    data = np.amax(x.data, axis=dim, keepdims=keepdim)
    requires_grad = x.requires_grad
    dependencies = []

    if x.requires_grad:
        def grad_fn(grad: np.ndarray, _) -> np.ndarray:
            if dim is None:
                ret = np.zeros_like(x.data)
                x_am = x.data.argmax()
                idx = np.unravel_index(x_am, x.shape)
                ret[idx] = grad
                return ret
            else:
                ret = np.zeros_like(x.data)
                x_am = x.data.argmax(axis=dim)
                x_am = np.expand_dims(x_am, axis=dim)
                grad_r = grad if keepdim else np.expand_dims(grad, axis=dim)
                np.put_along_axis(ret, x_am, grad_r, axis=dim)
                return ret

        dependencies.append(Dependency(x, grad_fn, meta={"name": "max"}))

    return Tensor(data, requires_grad, dependencies)


def min(x: Tensor, dim=None, keepdim=False) -> Tensor:
    # TODO: fix epsilon
    data = np.amin(x.data, axis=dim, keepdims=keepdim)
    requires_grad = x.requires_grad
    dependencies = []

    if x.requires_grad:
        def grad_fn(grad: np.ndarray, _) -> np.ndarray:
            if dim is None:
                ret = np.zeros_like(x.data)
                x_am = x.data.argmin()
                idx = np.unravel_index(x_am, x.shape)
                ret[idx] = grad
                return ret
            else:
                ret = np.zeros_like(x.data)
                x_am = x.data.argmin(axis=dim)
                x_am = np.expand_dims(x_am, axis=dim)
                grad_r = grad if keepdim else np.expand_dims(grad, axis=dim)
                np.put_along_axis(ret, x_am, grad_r, axis=dim)
                return ret

        dependencies.append(Dependency(x, grad_fn, meta={"name": "max"}))

    return Tensor(data, requires_grad, dependencies)


def log(x: Tensor) -> Tensor:
    data = np.log(x.data)
    requires_grad = x.requires_grad
    dependencies = []

    if x.requires_grad:
        def grad_fn(grad: np.ndarray, _) -> np.ndarray:
            return grad * (1 / x.data)

        dependencies.append(Dependency(x, grad_fn, meta={"name": "log"}))

    return Tensor(data, requires_grad, dependencies)


def exp(x: Tensor) -> Tensor:
    data = np.exp(x.data)
    requires_grad = x.requires_grad
    dependencies = []

    if x.requires_grad:
        def grad_fn(grad: np.ndarray, _) -> np.ndarray:
            return grad * data

        dependencies.append(Dependency(x, grad_fn, meta={"name": "exp"}))

    return Tensor(data, requires_grad, dependencies)


def dropout(x: Tensor, p: float = 0.5, training: bool = True) -> Tensor:
    if training:
        mask = np.random.rand(*x.shape) > p
        data = x.data * mask * (1 / (1 - p))
        requires_grad = x.requires_grad
        dependencies = []

        if x.requires_grad:
            def grad_fn(grad: np.ndarray, _) -> np.ndarray:
                return mask * grad / (1 - p)

            dependencies.append(Dependency(x, grad_fn, meta={"name": "dropout"}))

        return Tensor(data, requires_grad, dependencies)
    else:
        return x


def conv2d(x: Tensor, weight: Tensor,
           bias: Optional[Tensor] = None,
           stride: Union[Tuple, List, int] = 1,
           dilation: Union[Tuple, List, int] = 1,
           groups: int = 1) -> Tensor:
    """

    :param x: (bs, ch_i, h_i, w_i)
    :param weight: (ch_o, ch_i_g, h_k, w_k)
    :param bias: (ch_o)
    :param stride: -> (int, int)
    :param dilation: -> (int, int)
    :param groups: int
    :return (bs, ch_o, h_o, w_o)
    """
    assert x.dim() == 4
    bs, ch_i, h_i, w_i = x.shape
    ch_o, ch_i_g, h_k, w_k = weight.shape
    assert bias is None or bias.dim() == 1

    # make sure stride is (int, int)
    stride_err_msg = "value of `stride` must be tuple of 2 or int"
    if type(stride) == int:
        stride = (stride, stride)
    elif type(stride) in [tuple, list]:
        assert len(stride) == 2, stride_err_msg
    else:
        raise ValueError(stride_err_msg)

    # make sure dilation
    dia_err_msg = "value of `dilation` must be tuple of 2 or int"
    if type(dilation) == int:
        dilation = (dilation, dilation)
    elif type(dilation) in [tuple, int]:
        assert len(dilation) == 2, dia_err_msg
    else:
        raise ValueError(dia_err_msg)

    # make sure groups
    assert ch_i % groups == 0 and ch_o % groups == 0
    ch_o_g = ch_o // groups

    h_o = math.floor((h_i - dilation[0] * (h_k - 1) - 1) / stride[0] + 1)
    w_o = math.floor((w_i - dilation[1] * (w_k - 1) - 1) / stride[1] + 1)

    x_col = U.im2col2d(x.data, weight.shape, stride, dilation)
    x_col = x_col.reshape((bs, h_o * w_o, groups, ch_i_g * h_k * w_k))
    w_col = weight.data.reshape((groups, ch_o_g, ch_i_g * h_k * w_k))

    data = np.matmul(w_col, x_col, axes=[(1, 2), (3, 1), (2, 3)])
    data = data.reshape((bs, ch_o, h_o, w_o))
    if bias is not None:
        data += bias.data.reshape((1, ch_o, 1, 1))

    bias_requires_grad = bias is not None and bias.requires_grad
    requires_grad = x.requires_grad or weight.requires_grad or bias_requires_grad
    dependencies = []

    if x.requires_grad:
        def grad_fn_x(grad: np.ndarray, _) -> np.ndarray:
            w_rot = U.rot180_2d(weight.data)  # ch_o, ch_i_g, h_k, w_k
            w_rot = w_rot.transpose((1, 0, 2, 3))  # ch_i_g, ch_o, h_k, w_k
            w_rot_col = w_rot.reshape((ch_i_g, groups, ch_o_g * h_k * w_k))
            grad_p_s = U.pad_stride(grad, w_rot.shape, stride)
            grad_col = U.im2col2d(grad_p_s, w_rot.shape, stride=(1, 1), dilation=dilation)
            grad_col = grad_col.reshape((bs, h_i * w_i, groups, ch_o_g * h_k * w_k))
            # ch_i_g, groups, ch_o_g * h_k * w_k
            # bs, h_i * w_i, groups, ch_o_g * h_k * w_k
            grad_x = np.matmul(w_rot_col, grad_col, axes=[(0, 2), (3, 1), (2, 3)])
            grad_x = grad_x.reshape((bs, ch_i, h_i, w_i))
            assert grad_x.shape == x.shape
            return grad_x

        dependencies.append(Dependency(x, grad_fn_x, meta={"name": "conv2d_x"}))

    if weight.requires_grad:
        def grad_fn_w(grad: np.ndarray, _) -> np.ndarray:
            grad_p_s = U.unstride(grad, stride)
            _, _, h_o_p, w_o_p = grad_p_s.shape
            x_col_w = U.im2col2d(x.data, grad_p_s.shape, stride=(1, 1), dilation=dilation)  # bs, h_k*w_k, ch_i, h_o*w_o
            x_col_w = x_col_w.reshape((bs, h_k * w_k, groups, ch_i_g, h_o * w_o))
            x_col_w = x_col_w.transpose((3, 2, 1, 0, 4))  # ch_i_g, g, h_k*w_k, bs, h_o*w_o
            x_col_w = x_col_w.reshape((ch_i_g, groups, h_k * w_k, bs * h_o * w_o))
            grad_col = grad_p_s.transpose((1, 0, 2, 3))  # ch_o, bs, h_o, w_o
            grad_col = grad_col.reshape((groups, ch_o_g, bs * h_o_p * w_o_p))
            grad_w = np.matmul(grad_col, x_col_w, axes=[(1, 2), (3, 2), (2, 3)])  # ch_i_g, g, ch_o_g, h_k*w_k
            grad_w = grad_w.transpose((1, 2, 0, 3))  # g, ch_o_g, ch_i_g, h_k*w_k
            grad_w = grad_w.reshape((ch_o, ch_i_g, h_k, w_k))
            assert grad_w.shape == weight.shape  # (ch_o, ch_i_g, h_k, w_k)
            return grad_w

        dependencies.append(Dependency(weight, grad_fn_w, meta={"name": "conv2d_weight"}))

    if bias_requires_grad:
        def grad_fn_b(grad: np.ndarray, _) -> np.ndarray:
            grad = grad.reshape((bs, ch_o, h_o * w_o))
            grad_b = np.sum(grad, axis=2)
            grad_b = np.sum(grad_b, axis=0)
            assert grad_b.shape == bias.shape
            return grad_b

        dependencies.append(Dependency(bias, grad_fn_b, meta={"name": "conv2d_bias"}))

    return Tensor(data, requires_grad, dependencies)


def relu(x: Tensor) -> Tensor:
    data = np.maximum(x.data, 0)
    requires_grad = x.requires_grad
    dependencies = []

    if requires_grad:
        def grad_fn(grad: np.ndarray, _) -> np.ndarray:
            return grad * np.where(x.data > 0, 1, 0)

        dependencies.append(Dependency(x, grad_fn, meta={"name": "relu"}))

    return Tensor(data, requires_grad, dependencies)


def sigmoid(x: Tensor) -> Tensor:
    data = 1 / (1 + np.exp(-x.data))
    requires_grad = x.requires_grad
    dependencies = []

    if requires_grad:
        def grad_fn(grad: np.ndarray, _) -> np.ndarray:
            return grad * data * (1 - data)

        dependencies.append(Dependency(x, grad_fn, meta={"name": "sigmoid"}))

    return Tensor(data, requires_grad, dependencies)


def softmax(x: Tensor, dim=-1) -> Tensor:
    e = np.exp(x.data - np.max(x.data))
    data = e / np.sum(e, axis=dim, keepdims=True)  # (bs, n)
    requires_grad = x.requires_grad
    dependencies = []

    if requires_grad:
        def grad_fn(grad: np.ndarray, _) -> np.ndarray:
            s = data[:, :, np.newaxis]  # (bs, n, 1)
            a = np.array([np.diagflat(s[i, :, :]) for i in range(s.shape[0])])  # (bs, n, n)
            b = np.matmul(s, s.transpose((0, 2, 1)))  # (bs, n, n)
            c = grad[:, np.newaxis, :]
            d = np.matmul(c, (a - b))
            return d.squeeze()

        dependencies.append(Dependency(x, grad_fn, meta={"name": "softmax"}))

    return Tensor(data, requires_grad, dependencies)


def tanh(x: Tensor) -> Tensor:
    data = np.tanh(x.data)
    requires_grad = x.requires_grad
    dependencies = []

    if requires_grad:
        def grad_fn(grad: np.ndarray, _) -> np.ndarray:
            return grad * (1 - data * data)

        dependencies.append(Dependency(x, grad_fn, meta={"name": "tanh"}))

    return Tensor(data, requires_grad, dependencies)


def nll(x: Tensor, y: Tensor, w: Optional[Tensor] = None, dim=-1) -> Tensor:
    indices = np.expand_dims(y.data, axis=-1)
    data = np.take_along_axis(x.data, indices, axis=dim)
    if w:
        data = np.take_along_axis(w, indices, axis=0) * data
    data = data.squeeze()
    requires_grad = x.requires_grad
    dependencies = []

    if x.requires_grad:
        def grad_fn(grad: np.ndarray, _) -> np.ndarray:
            assert grad.shape == data.shape
            ret = np.zeros_like(x.data)
            np.put_along_axis(ret, indices, np.expand_dims(grad, -1), axis=dim)
            assert ret.shape == x.shape
            return ret

        dependencies.append(Dependency(x, grad_fn, meta={"name": "nll"}))

    return Tensor(data, requires_grad, dependencies)
